@@ -17,6 +17,13 @@ except ImportError:
     HAVE_BULLET = False
 from direct.stdpy.file import open # pylint: disable=redefined-builtin
 
+if LVector3 is LVector3f:
+    CPTA_stdfloat = CPTA_float
+    PTA_stdfloat = PTA_float
+else:
+    CPTA_stdfloat = CPTA_double
+    PTA_stdfloat = PTA_double
+
 load_prc_file_data(
     __file__,
     'interpolate-frames #t\n'
@@ -97,6 +104,7 @@ class Converter():
         self.node_paths = {}
         self.scenes = {}
         self.skeletons = {}
+        self.joint_parents = {}
 
         # Coordinate system transform matrix
         self.csxform = LMatrix4.convert_mat(CS_yup_right, CS_default)
@@ -167,7 +175,9 @@ class Converter():
 
             node_name = gltf_node.get('name', 'node'+str(nodeid))
             if nodeid in self._joint_nodes:
-                # don't handle joints here
+                # Handle non-joint children of joints, but don't add joints themselves
+                for child_nodeid in gltf_node.get('children', []):
+                    add_node(root, gltf_scene, child_nodeid, jvtmap, cvsmap)
                 return
 
             jvtmap = dict(jvtmap)
@@ -260,15 +270,27 @@ class Converter():
                     if isinstance(light, Light):
                         root.set_light(lnp)
 
-                if 'BLENDER_physics' in gltf_node['extensions']:
-                    gltf_collisions = gltf_node['extensions']['BLENDER_physics']
-                    gltf_rigidbody = gltf_node['extensions']['BLENDER_physics']
-                    collision_shape = gltf_collisions['collisionShapes'][0]
-                    shape_type = collision_shape['shapeType']
+                has_physics = (
+                    'BLENDER_physics' in gltf_node['extensions'] or
+                    'PANDA3D_physics_collision_shapes' in gltf_node['extensions']
+                )
+                if has_physics:
+                    gltf_collisions = gltf_node['extensions'].get(
+                        'PANDA3D_physics_collision_shapes',
+                        gltf_node['extensions']['BLENDER_physics']
+                    )
+                    gltf_rigidbody = gltf_node['extensions'].get('BLENDER_physics', None)
+                    if 'PANDA3D_physics_collision_shapes' in gltf_node['extensions']:
+                        collision_shape = gltf_collisions['shapes'][0]
+                        shape_type = collision_shape['type']
+                    else:
+                        collision_shape = gltf_collisions['collisionShapes'][0]
+                        shape_type = collision_shape['shapeType']
                     bounding_box = collision_shape['boundingBox']
                     radius = max(bounding_box[0], bounding_box[1]) / 2.0
                     height = bounding_box[2]
                     geomnode = None
+                    intangible = gltf_collisions.get('intangible', False)
                     if 'mesh' in collision_shape:
                         try:
                             geomnode = self.meshes[collision_shape['mesh']]
@@ -297,6 +319,7 @@ class Converter():
                             bounding_box,
                             radius,
                             height,
+                            intangible,
                             gltf_rigidbody
                         )
                     else:
@@ -307,7 +330,7 @@ class Converter():
                             bounding_box,
                             radius,
                             height,
-                            gltf_rigidbody
+                            intangible
                         )
                     if phynode is not None:
                         np.attach_new_node(phynode)
@@ -344,6 +367,13 @@ class Converter():
                     tmp.set_attrib(CullFaceAttrib.make_reverse())
                     geomnode.reparent_to(tmp)
 
+            # Handle parenting to joints
+            joint = self.joint_parents.get(nodeid)
+            if joint:
+                xformnp = root.attach_new_node(PandaNode('{}-parent'.format(node_name)))
+                np.reparent_to(xformnp)
+                joint.add_net_transform(xformnp.node())
+
         for sceneid, gltf_scene in enumerate(gltf_data.get('scenes', [])):
             scene_name = gltf_scene.get('name', 'scene'+str(sceneid))
             scene_root = NodePath(ModelRoot(scene_name))
@@ -776,15 +806,21 @@ class Converter():
                 uvs = uv_data.get_data2f()
                 uv_data.set_data2f(uvs[0], 1 - uvs[1])
 
-        # Flip morph deltas from Y-up to Z-up.  This is apparently not done by
-        # transform_vertices(), below, so we do it ourselves.
         if self.compose_cs == CS_yup_right:
+            # Flip morph deltas from Y-up to Z-up.  This is apparently not done by
+            # transform_vertices(), below, so we do it ourselves.
             for morph_i in range(reg_format.get_num_morphs()):
                 delta_data = GeomVertexRewriter(vdata, reg_format.get_morph_delta(morph_i))
 
                 while not delta_data.is_at_end():
                     data = delta_data.get_data3f()
                     delta_data.set_data3f(data[0], -data[2], data[1])
+            # Flip tangents from Y-up to Z-up.
+            if 'TANGENT' in mesh_attribs:
+                tangent = GeomVertexRewriter(vdata, InternalName.make('tangent'))
+                while not tangent.is_at_end():
+                    data = tangent.get_data4f()
+                    tangent.set_data4f(data[0], -data[2], data[1], data[3])
 
         # Repack mesh data
         vformat = GeomVertexFormat()
@@ -847,6 +883,10 @@ class Converter():
             end = start + index_acc['count'] * buffview.get('byteStride', 1) * prim.index_stride
             handle.copy_data_from(buff[start:end])
             handle = None
+        else:
+            index_acc = gltf_data['accessors'][gltf_primitive['attributes']["POSITION"]]
+            start = index_acc.get('byteOffset', 0)
+            prim.setNonindexedVertices(start, index_acc['count'])
 
         # Assign a material
         matid = gltf_primitive.get('material', None)
@@ -909,7 +949,7 @@ class Converter():
             if denom != 0.0:
                 fconst = 1.0 / denom
                 tangent = (edge1.xyz * duv2.y - edge2.xyz * duv1.y) * fconst
-                bitangent = (edge1.xyz * duv2.x - edge2.xyz * duv1.x) * fconst
+                bitangent = (edge2.xyz * duv1.x - edge1.xyz * duv2.x) * fconst
             else:
                 tangent = LVector3(0)
                 bitangent = LVector3(0)
@@ -929,7 +969,10 @@ class Converter():
                 tangent.z,
                 -1.0 if normal.cross(tan0).dot(tan1) < 0 else 1.0
             )
-            tangent_writer.set_data4(tangent4)
+            if self.compose_cs == CS_yup_right:
+                tangent_writer.set_data4(tangent4[0], -tangent4[2], tangent4[1], tangent4[3])
+            else:
+                tangent_writer.set_data4(tangent4)
 
         geom.set_vertex_data(gvd)
 
@@ -1003,7 +1046,8 @@ class Converter():
                 for idx in range(start, end, 4)
             ]
             num_frames = time_acc['count']
-            fps = num_frames / time_data[-1]
+            end_time = time_data[-1]
+            fps = num_frames / time_data[-1] if end_time != 0 else 24
 
             bundle_name = anim_name
             bundle = AnimBundle(bundle_name, fps, num_frames)
@@ -1118,6 +1162,11 @@ class Converter():
         def create_joint(parent, nodeid, transform):
             node = gltf_data['nodes'][nodeid]
             node_name = node.get('name', 'bone'+str(nodeid))
+
+            if 'mesh' in node:
+                self.joint_parents[nodeid] = parent
+                return
+
             inv_transform = LMatrix4(transform)
             inv_transform.invert_in_place()
             joint_index = None
@@ -1309,19 +1358,22 @@ class Converter():
             scale_vals[2] = scale_vals[2][:1]
 
         # Write data to tables
-        group.set_table(b'x', CPTAFloat(PTAFloat(loc_vals[0])))
-        group.set_table(b'y', CPTAFloat(PTAFloat(loc_vals[1])))
-        group.set_table(b'z', CPTAFloat(PTAFloat(loc_vals[2])))
+        group.set_table(b'x', CPTA_stdfloat(PTA_stdfloat(loc_vals[0])))
+        group.set_table(b'y', CPTA_stdfloat(PTA_stdfloat(loc_vals[1])))
+        group.set_table(b'z', CPTA_stdfloat(PTA_stdfloat(loc_vals[2])))
 
-        group.set_table(b'h', CPTAFloat(PTAFloat(rot_vals[0])))
-        group.set_table(b'p', CPTAFloat(PTAFloat(rot_vals[1])))
-        group.set_table(b'r', CPTAFloat(PTAFloat(rot_vals[2])))
+        group.set_table(b'h', CPTA_stdfloat(PTA_stdfloat(rot_vals[0])))
+        group.set_table(b'p', CPTA_stdfloat(PTA_stdfloat(rot_vals[1])))
+        group.set_table(b'r', CPTA_stdfloat(PTA_stdfloat(rot_vals[2])))
 
-        group.set_table(b'i', CPTAFloat(PTAFloat(scale_vals[0])))
-        group.set_table(b'j', CPTAFloat(PTAFloat(scale_vals[1])))
-        group.set_table(b'k', CPTAFloat(PTAFloat(scale_vals[2])))
+        group.set_table(b'i', CPTA_stdfloat(PTA_stdfloat(scale_vals[0])))
+        group.set_table(b'j', CPTA_stdfloat(PTA_stdfloat(scale_vals[1])))
+        group.set_table(b'k', CPTA_stdfloat(PTA_stdfloat(scale_vals[2])))
 
         for childid in bone.get('children', []):
+            gltf_node = gltf_data['nodes'][childid]
+            if 'mesh' in gltf_node:
+                continue
             self.build_animation_skeleton(character, group, childid, num_frames, gltf_anim, gltf_data)
 
     def build_animation_morph(self, parent, nodeid, num_frames, gltf_anim, gltf_data, recurse=True):
@@ -1471,9 +1523,9 @@ class Converter():
 
         self.lights[lightid] = node
 
-    def load_physics_bullet(self, node_name, geomnode, shape_type, bounding_box, radius, height, gltf_rigidbody):
+    def load_physics_bullet(self, node_name, geomnode, shape_type, bounding_box, radius, height, intangible, gltf_rigidbody): # pylint: disable=line-too-long
         shape = None
-        static = 'static' in gltf_rigidbody and gltf_rigidbody['static']
+        static = gltf_rigidbody is not None and 'static' in gltf_rigidbody and gltf_rigidbody['static']
 
         if shape_type == 'BOX':
             shape = bullet.BulletBoxShape(LVector3(*bounding_box) / 2.0)
@@ -1498,24 +1550,30 @@ class Converter():
                     mesh.add_geom(geom)
                 shape = bullet.BulletTriangleMeshShape(mesh, dynamic=not static)
         else:
-            print("Unknown collision shape ({}) for object ({})".format(shape_type, nodeid))
+            print("Unknown collision shape ({}) for object ({})".format(shape_type, node_name))
 
         if shape is not None:
-            phynode = bullet.BulletRigidBodyNode(node_name)
+            if intangible:
+                phynode = bullet.BulletGhostNode(node_name)
+            else:
+                phynode = bullet.BulletRigidBodyNode(node_name)
             phynode.add_shape(shape)
             if not static:
-                phynode.set_mass(gltf_rigidbody['mass'])
+                mass = 1.0 if gltf_rigidbody is None else gltf_rigidbody.get('mass', 1.0)
+                phynode.set_mass(mass)
             return phynode
         else:
-            print("Could not create collision shape for object ({})".format(nodeid))
+            print("Could not create collision shape for object ({})".format(node_name))
 
-    def load_physics_builtin(self, node_name, geomnode, shape_type, bounding_box, radius, height, _gltf_rigidbody):
+    def load_physics_builtin(self, node_name, geomnode, shape_type, bounding_box, radius, height, intangible):
         phynode = CollisionNode(node_name)
 
+        solids = []
+
         if shape_type == 'BOX':
-            phynode.add_solid(CollisionBox(Point3(0, 0, 0), *LVector3(*bounding_box) / 2.0))
+            solids.append(CollisionBox(Point3(0, 0, 0), *LVector3(*bounding_box) / 2.0))
         elif shape_type == 'SPHERE':
-            phynode.add_solid(CollisionSphere(0, 0, 0, radius))
+            solids.append(CollisionSphere(0, 0, 0, radius))
         elif shape_type in ('CAPSULE', 'CYLINDER', 'CONE'):
             if shape_type != 'CAPSULE':
                 print(
@@ -1527,7 +1585,7 @@ class Converter():
             half_height = height / 2.0 - radius
             start = LPoint3(0, 0, -half_height)
             end = LPoint3(0, 0, half_height)
-            phynode.add_solid(CollisionCapsule(start, end, radius))
+            solids.append(CollisionCapsule(start, end, radius))
         elif shape_type in ('MESH', 'CONVEX_HULL'):
             if shape_type != 'MESH':
                 print(
@@ -1549,14 +1607,19 @@ class Converter():
 
                 polys = zip(*([iter(verts)] * 3))
                 for poly in polys:
-                    phynode.add_solid(CollisionPolygon(*poly))
+                    solids.append(CollisionPolygon(*poly))
         else:
             print("Unknown collision shape ({}) for object ({})".format(shape_type, node_name))
 
+        for solid in solids:
+            if intangible:
+                solid.set_tangible(False)
+            phynode.add_solid(solid)
+
         if phynode.solids:
             return phynode
         else:
-            print("Could not create collision shape for object ({})".format(nodeid))
+            print("Could not create collision shape for object ({})".format(node_name))
 
 
 def read_glb_chunk(glb_file):
